#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 sequence label decoder

Authors: PanXu
Date:    2020/07/05 17:09:00
"""
import torch

from easytext.tests import ASSERT

from easytext.data import LabelVocabulary
from easytext.label_decoder import SequenceLabelDecoder


def test_sequence_label_decoder():
    """
    测试 sequence label decoder
    :return:
    """
    sequence_label_list = list()
    expect_spans = list()

    sequence_label = ["B-T", "I-T", "O-T"]
    expect = [{"label": "T", "begin": 0, "end": 2}]

    sequence_label_list.append(sequence_label)
    expect_spans.append(expect)

    sequence_label = ["B-T", "I-T", "I-T"]
    expect = [{"label": "T", "begin": 0, "end": 3}]

    sequence_label_list.append(sequence_label)
    expect_spans.append(expect)

    sequence_label = ["B-T", "I-T", "I-T", "B-T"]
    expect = [{"label": "T", "begin": 0, "end": 3},
              {"label": "T", "begin": 3, "end": 4}]

    sequence_label_list.append(sequence_label)
    expect_spans.append(expect)

    label_vocabulary = LabelVocabulary(sequence_label_list,
                                       padding=LabelVocabulary.PADDING)

    sequence_label_indices = list()
    mask_list = list()

    max_sequence_len = 4
    for sequence_labels in sequence_label_list:
        sequence_label_index = [label_vocabulary.index(label) for label in sequence_labels]

        mask = [1] * len(sequence_label_index) + [0] * (max_sequence_len - len(sequence_label_index))

        sequence_label_index.extend([label_vocabulary.padding_index] * (max_sequence_len - len(sequence_label_index)))

        sequence_label_indices.append(sequence_label_index)
        mask_list.append(mask)

    sequence_label_indices = torch.tensor(sequence_label_indices, dtype=torch.long)
    mask = torch.tensor(mask_list, dtype=torch.uint8)

    decoder = SequenceLabelDecoder(label_vocabulary=label_vocabulary)

    spans = decoder(label_indices=sequence_label_indices, mask=mask)

    ASSERT.assertListEqual(expect_spans, spans)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
英文tokenizer

Authors: panxu(panxu@baidu.com)
Date:    2020/06/09 11:18:00
"""
from typing import List

from .token import Token
from .tokenizer import Tokenizer


class EnTokenizer(Tokenizer):
    """
    英文tokenizer, 会对文本进行分隔成单词，再将单词全部转换成小写，作为token.
    默认的分隔符是 " "
    """

    def __init__(self, is_remove_invalidate_char:bool = False, separator: str = " "):
        """
        初始化
        :param is_remove_invalidate_char: 是否移除无效的字符，注意在实体识别的时候要小心，因为移除可能导致label无法对其。
        :param separator: 分隔符，用来切分英文文本成单词，默认是 " "
        """
        super().__init__(is_remove_invalidate_char=is_remove_invalidate_char)
        self._separator = separator

    def _tokenize(self, text: str) -> List[Token]:

        return [Token(word.lower()) for word in text.split(self._separator)]

    def batch_tokenize(self, batch_text: List[str]) -> List[List[Token]]:

        return [self._tokenize(text) for text in batch_text]


#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
label index decoder

Authors: PanXu
Date:    2020/07/05 11:12:00
"""

from typing import Tuple
import torch


class LabelIndexDecoder:
    """
    将 logits 解码成 label index
    """

    def __call__(self,
                 logits: torch.Tensor,
                 mask: torch.ByteTensor) -> torch.LongTensor:
        """
        将 logits 解码成 label index。
        例如:
        logits = [[0.3, 0.7], [0.8, 0.2]]
        返回的 index = [1, 0]

        :param logits: 模型输出 logits
        :param mask: mask
        :return: index
        """
        raise NotImplementedError()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
config 工厂

Authors: panxu(panxu@baidu.com)
Date:    2020/06/25 18:57:00
"""

from typing import Dict


class ConfigFactory:
    """
    config 创建工厂
    """

    def create(self) -> Dict:
        """
        创建 config
        """
        raise NotImplementedError()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
f1 metric 基类，具体的 f1 需要子类实现

Authors: panxu(panxu@baidu.com)
Date:    2020/06/16 09:08:00
"""

from typing import Dict, Set
from collections import defaultdict

import torch

from .metric import Metric


class F1Metric(Metric):
    """
    f1 metric 基类，具体的 f1 需要子类实现，来完成
    _true_positives, _false_positives, _false_negatives
    三个字典的填充, 但是注意: 对于 *-overall 是会自动计算的，不需要
    指定计算。
    """

    PRECISION = "precision"
    RECALL = "recall"
    F1 = "f1"

    PRECISION_OVERALL = f"{PRECISION}-overall"
    RECALL_OVERALL = f"{RECALL}-overall"
    F1_OVERALL = f"{F1}-overall"

    def __init__(self) -> None:
        """
        初始化
        """
        super().__init__()

        # 下面之所以是字典，是为了计算多个 tag 的 f1
        self._true_positives: Dict[str, int] = defaultdict(int)
        self._false_positives: Dict[str, int] = defaultdict(int)
        self._false_negatives: Dict[str, int] = defaultdict(int)

    def __call__(self,
                 prediction_labels: torch.Tensor,
                 gold_labels: torch.Tensor,
                 mask: torch.LongTensor) -> Dict:
        """
        计算 metric. 返回的是 F1 字典:

        {"precision_[tag]": [value],
         "recall_[tag]" : [value],
         "f1-measure_[tag]": [value],
         "precision-overall": [value],
         "recall-overall": [value],
         "f1-measure-overall": [value]}

         其中的 [tag] 是 实际返回的tag

        :param prediction_labels: 预测的结果, shape: (B, SeqLen)
        :param gold_labels: 实际的结果, shape: (B, SeqLen)
        :param mask: 对 predictions 和 gold label 的 mask, shape: (B, SeqLen)
        :return: 当前的 metric 计算字典结果.
        """

        raise NotImplementedError()

    def _metric(self,
                true_positives: Dict[str, int],
                false_positives: Dict[str, int],
                false_negatives: Dict[str, int]) -> Dict:
        """
        计算 metric, 注意是 输入是的字典
        :param true_positives:
        :param false_positives:
        :param false_negatives:
        :return:
        """

        # 更新 累积 _true_positives
        for k, v in true_positives.items():
            self._true_positives[k] += v

        # 更新 累积 _false_positives
        for k, v in false_positives.items():
            self._false_positives[k] += v

        # 更新 累积 _false_negatives
        for k, v in false_negatives.items():
            self._false_negatives[k] += v

        all_tags: Set[str] = set()
        all_tags.update(true_positives.keys())
        all_tags.update(false_positives.keys())
        all_tags.update(false_negatives.keys())
        all_metrics = {}

        for tag in all_tags:
            precision, recall, f1_measure = self._compute_metrics(true_positives[tag],
                                                                  false_positives[tag],
                                                                  false_negatives[tag])
            precision_key = f"{F1Metric.PRECISION}-{tag}"
            recall_key = f"{F1Metric.RECALL}-{tag}"
            f1_key = f"{F1Metric.F1}-{tag}"
            all_metrics[precision_key] = precision
            all_metrics[recall_key] = recall
            all_metrics[f1_key] = f1_measure

        # Compute the precision, recall and f1 for overall
        precision, recall, f1_measure = self._compute_metrics(sum(true_positives.values()),
                                                              sum(false_positives.values()),
                                                              sum(false_negatives.values()))
        all_metrics[F1Metric.PRECISION_OVERALL] = precision
        all_metrics[F1Metric.RECALL_OVERALL] = recall
        all_metrics[F1Metric.F1_OVERALL] = f1_measure

        return all_metrics

    @property
    def metric(self) -> Dict:
        """
        :return

        一个包含所有metric的字典:
        {
            precision-[tag] : float
            recall-[tag] : float
            f1-[tag] : float
            ...
        }

        另外注意 precision-overall, recall-overall, f1-overall 是所有的综合指标, 这是非常有必要的
        因为有多个tag，那么模型的指标衡量需要一个综合指标来衡量。
        """
        return self._metric(true_positives=self._true_positives,
                            false_positives=self._false_positives,
                            false_negatives=self._false_negatives)

    @staticmethod
    def _compute_metrics(true_positives: int, false_positives: int, false_negatives: int):
        """
        计算 f1 metric
        """
        precision = float(true_positives) / float(true_positives + false_positives + 1e-13)
        recall = float(true_positives) / float(true_positives + false_negatives + 1e-13)
        f1_measure = 2. * ((precision * recall) / (precision + recall + 1e-13))
        return precision, recall, f1_measure

    def reset(self):
        """
        将所有的状态reset, f1 重新计算。
        """
        self._true_positives = defaultdict(int)
        self._false_positives = defaultdict(int)
        self._false_negatives = defaultdict(int)
        return self




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
指标

Authors: panxu(panxu@baidu.com)
Date:    2020/05/16 00:57:00
"""

from .metric import Metric
from .metric import ModelMetricAdapter
from .metric import ModelTargetMetric

from .f1_metric import F1Metric
from .span_f1_metric import SpanF1Metric
from .acc_metric import AccMetric
from .label_f1_metric import LabelF1Metric





#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
预训练词向量载入器

Authors: panxu(panxu@baidu.com)
Date:    2020/06/24 11:35:00
"""

from .pretrained_word_embedding_loader import PretrainedWordEmbeddingLoader
from .general_pretrained_word_embedding_loader import GeneralPretrainedWordEmbeddingLoader
from .glove import GloveLoader



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
brief

Authors: panxu(panxu@baidu.com)
Date:    2020/05/16 00:33:00
"""
from .metric_tracker import MetricTracker
from .trainer import Trainer
from .config_factory import ConfigFactory




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
brief

Authors: panxu(panxu@baidu.com)
Date:    2020/05/15 10:50:00
"""
import os
from unittest import TestCase

ASSERT = TestCase()

# 设置 root， 放在 tests 中，是因为 easytext 是库，只有在 tests 的时候，涉及到数据存储
ROOT_PATH = os.path.join(os.path.dirname(__file__), "../..")



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
max label decoder

Authors: PanXu
Date:    2020/07/06 10:13:00
"""
from typing import Union
import torch

from .label_index_decoder import LabelIndexDecoder


class MaxLabelIndexDecoder(LabelIndexDecoder):
    """
    max label index decoder
    """

    def __call__(self, logits: torch.Tensor, mask: Union[None, torch.ByteTensor] = None) -> torch.LongTensor:
        """
        基于 argmax 从 logits 中提取出 max index
        :param logits: logits shape: (batch_size, num_label)
        :param mask: mask 必须是 None
        :return:
        """

        if mask is not None:
            raise RuntimeError(f"mask 必须是 None")

        return torch.argmax(logits, dim=-1)




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
实例

Authors: panxu(panxu@baidu.com)
Date:    2020/05/13 15:04:00
"""

from collections import OrderedDict


class Instance(OrderedDict):
    """
    Instance 类型
    """

    pass


#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
训练器

Authors: panxu(panxu@baidu.com)
Date:    2020/05/16 00:34:00
"""

import os
import torch
import logging
from typing import List, Union
from tqdm import tqdm
import shutil

from torch.utils.data import DataLoader
from torch.optim.lr_scheduler import ReduceLROnPlateau

from easytext.model import Model
from easytext.loss import Loss
from easytext.optimizer import OptimizerFactory
from easytext.optimizer import LRSchedulerFactory
from easytext.data.model_collate import ModelInputs
from easytext.metrics import ModelMetricAdapter
from easytext.utils.json_util import json2str
from easytext.utils.nn import cuda_util
from easytext.trainer.metric_tracker import MetricTracker


class Trainer:
    """
    训练器
    """
    _TRAIN = 0
    _EVALUATE = 1

    def __init__(self,
                 serialize_dir: str,
                 num_epoch: int,
                 model: Model,
                 loss: Loss,
                 metrics: ModelMetricAdapter,
                 optimizer_factory: OptimizerFactory,
                 lr_scheduler_factory: LRSchedulerFactory = None,
                 patient: int = None,
                 num_check_point_keep: int = None,
                 cuda_devices: List[str] = None):
        """
        训练器初始化
        :param num_epoch: 训练的 epoch 数量
        :param model: 要训练的模型
        :param loss: 模型的 loss function
        :param metrics: 模型的指标计算
        :param optimizer_factory: 模型的优化器的创建工厂。为什么不直接使用优化器？是因为, 优化器的创建依赖于 model, 所以
        这里传递的参数 optimizer factory, 避免使用者在 trainer 外面生成 optimizer, 导致在 trainer 内 optimizer 依赖于
        model 的参数产生问题。典型问题是: 设置 cuda.
        :param serialize_dir: 训练存储的文件路径
        :param patient: early stopping 的 patient. 如果是 `None`, 将不会进行 early stopping;
        否则, 当前训练的指标超出了 patient 个 epoch 将会 early stopping.
        :param num_check_point_keep: checkpoint 保留的数量。如果是 `None` 则全部保留;
        否则，保留 num_check_point_keep 个checkpoint.
        :param cuda_devices: cuda device列表. 字符串类型，那么就是 "cuda:0" 这种格式。
        """

        if cuda_devices is not None and len(cuda_devices) != 1:
            raise RuntimeError(f"目前仅仅支持单卡训练, 设置的 cuda devices 是 {cuda_devices}")

        if cuda_devices is not None:
            self._cuda_devices = [torch.device(device) for device in cuda_devices]
            self._model = model.cuda(self._cuda_devices[0])
        else:
            self._cuda_devices = None
            self._model = model

        self._loss = loss
        self._metrics = metrics
        self._optimizer = optimizer_factory.create(model=self._model)

        if lr_scheduler_factory is not None:
            self._lr_scheduler = lr_scheduler_factory.create(optimizer=self.optimizer,
                                                             model=self.model)
        else:
            self._lr_scheduler = None

        self._serialize_dir = serialize_dir
        self._metric_tracker = MetricTracker(patient=patient)
        self._num_check_point_keep = num_check_point_keep
        self._num_epoch = num_epoch
        self._current_epoch: int = None

    @property
    def model(self):
        return self._model

    @property
    def loss(self):
        return self._loss

    @property
    def metrics(self):
        return self._metrics

    @property
    def metric_tracker(self):
        return self._metric_tracker

    @property
    def optimizer(self):
        return self._optimizer

    @property
    def serialize_dir(self):
        return self._serialize_dir

    @property
    def num_epoch(self):
        return self._num_epoch

    @property
    def current_epoch(self):
        return self._current_epoch

    def save_checkpoint(self,
                        epoch: int) -> "Trainer":
        """
        保存 checkpoint 到指定的路径下。
        :param serialize_dir: 保存 模型训练相关的文件夹，包括 checkpoint 以及 其他信息
        :param epoch: 保存的 epoch
        :return: self
        """

        # 创建 checkpoint dir

        saved_dir = os.path.join(self._serialize_dir, f"checkpoint_epoch_{epoch}")

        if not os.path.isdir(saved_dir):
            os.makedirs(saved_dir)

        # 模型保存
        model_file_path = os.path.join(saved_dir, "model.pt")
        torch.save(self._model.state_dict(), model_file_path)

        # 优化器保存
        optimizer_file_path = os.path.join(saved_dir, "optimizer.pt")
        torch.save(self._optimizer.state_dict(), optimizer_file_path)

        # 保存lr scheduler

        lr_scheduler_file_path = os.path.join(saved_dir, "lr_scheduler.pt")

        if self._lr_scheduler is None:
            torch.save(None, lr_scheduler_file_path)
        else:
            torch.save(self._lr_scheduler.sate_dict(), lr_scheduler_file_path)

        # metric 保存
        # 这里保存的是当前 epoch 的 metric, 在最外面还会保存一份完整的 metric tracker
        metric_file_path = os.path.join(saved_dir, "metric.json")

        metric = self._metric_tracker[epoch]
        with open(metric_file_path, mode="w", encoding="utf-8") as f:
            f.write(f"{json2str(metric, indent=2)}\n")

        # metric tracker 保存
        metric_tracker_file_path = os.path.join(self._serialize_dir, "metric_tracker.json")
        self._metric_tracker.save(metric_tracker_file_path)

        # save best
        if self._metric_tracker.best().epoch == epoch:
            # 保存当前为best
            best_dir = os.path.join(self._serialize_dir, "best")

            if os.path.isdir(best_dir):
                # 已经存在，移动到 backup
                best_bak_dir = os.path.join(self._serialize_dir, "best_bak")

                if os.path.isdir(best_bak_dir):  # 删除 bak
                    shutil.rmtree(best_bak_dir)

                # 将 best 的 备份到 best_bak
                shutil.move(best_dir, best_bak_dir)

            shutil.copytree(saved_dir, best_dir)

        # 删除keep last之外的
        if self._num_check_point_keep is not None:
            for removed_epoch in range(1, epoch - self._num_check_point_keep + 1):
                removed_epoch_dir = os.path.join(self._serialize_dir, f"checkpoint_epoch_{removed_epoch}")

                if os.path.isdir(removed_epoch_dir):
                    shutil.rmtree(removed_epoch_dir)

        return self

    @staticmethod
    def _find_last_epoch(serialize_dir: str):
        """
        寻找最后的 epoch
        :param serialize_dir: serialize 目录
        :return:
        """
        # 找到 last epoch
        last_epoch = None
        for file_name in os.listdir(serialize_dir):

            dir_path = os.path.join(serialize_dir, file_name)

            if os.path.isdir(dir_path):
                parts = file_name.split("_")
                if len(parts) == 3 and parts[0] == "checkpoint" and parts[1] == "epoch":
                    epoch = int(parts[2])

                    if last_epoch is None:
                        last_epoch = epoch
                    else:
                        if epoch > last_epoch:
                            last_epoch = epoch
        return last_epoch

    def load_checkpoint(self,
                        serialize_dir: str) -> "Trainer":
        """
        载入 check point
        :param serialize_dir: 保存的路径
        :return: self
        """

        last_epoch = Trainer._find_last_epoch(serialize_dir=serialize_dir)

        if last_epoch is not None:
            self._current_epoch = last_epoch
            logging.info(f"Load checkpoint, 当前 epoch: {last_epoch}")
            saved_dir = os.path.join(serialize_dir, f"checkpoint_epoch_{last_epoch}")

            model_file_path = os.path.join(saved_dir, "model.pt")
            self._model.load_state_dict(torch.load(model_file_path))

            print(f"last epoch{last_epoch}, loaded: {self._model.state_dict()}")

            optimizer_file_path = os.path.join(saved_dir, "optimizer.pt")
            self._optimizer.load_state_dict(torch.load(optimizer_file_path))

            lr_scheduler_file_path = os.path.join(saved_dir, "lr_scheduler.pt")

            lr_state_dict = torch.load(lr_scheduler_file_path)
            if lr_state_dict is None:
                self._lr_scheduler = None
            else:
                self._lr_scheduler.load_state_dict(lr_state_dict)

            metric_tracker_file_path = os.path.join(serialize_dir, "metric_tracker.json")
            self._metric_tracker = MetricTracker.from_file(metric_tracker_file_path)
        else:
            raise RuntimeError(f"最后保存的epoch数据没有在 {self._serialize_dir} 中找到!")

        return self

    def _train_or_evaluate(self,
                           phrase: int,
                           data_loader: DataLoader) -> float:

        total_loss = 0.
        total_num = 0
        self._metrics.reset()

        if phrase == Trainer._TRAIN:
            self._model.train()
        elif phrase == Trainer._EVALUATE:
            self._model.eval()
        else:
            raise RuntimeError(f"phrase: {phrase} 应该是 {Trainer._TRAIN} 或 {Trainer._EVALUATE}")

        with torch.set_grad_enabled(phrase == Trainer._TRAIN):
            for model_inputs in tqdm(data_loader):
                model_inputs: ModelInputs = model_inputs

                batch_size, batch_inputs, labels = model_inputs.batch_size, \
                                                   model_inputs.model_inputs, \
                                                   model_inputs.labels

                # 设置到 cuda 训练
                if self._cuda_devices is not None:
                    batch_inputs = cuda_util.cuda(batch_inputs, cuda_device=self._cuda_devices[0])
                    labels = cuda_util.cuda(labels, cuda_device=self._cuda_devices[0])

                outputs = self._model(**batch_inputs)
                batch_loss: torch.Tensor = self._loss(outputs, labels)

                if phrase == Trainer._TRAIN:
                    self._optimizer.zero_grad()
                    batch_loss.backward()
                    self._optimizer.step()

                total_loss += batch_loss.detach().cpu().item() * batch_size
                total_num += batch_size

                batch_metrics, target_metric = self._metrics(model_outputs=outputs, golden_labels=labels)
                logging.info(f"Epoch: {self._current_epoch}, batch loss: {batch_loss},"
                             f"batch metrics: {json2str(batch_metrics)}, "
                             f"target metric: {json2str(target_metric)}")

        # total_loss = total_loss / total_num 这是合理的 loss, 因为所有的 total_num 是一样的所以，没有一般要再除以一次了
        return total_loss

    def recovery_train(self,
                       train_data_loader: DataLoader,
                       validation_data_loader: DataLoader):
        """
        恢复训练，是指从上次中断的位置重新开始训练
        :param train_data_loader: 训练数据集
        :param validation_data_loader:
        :return:
        """

        self.load_checkpoint(self.serialize_dir)
        self._train(train_data_loader=train_data_loader,
                    validation_data_loader=validation_data_loader)

    def evaluate(self,
                 validation_data_loader: DataLoader) -> float:
        """
        评估验证集
        :param validation_data_loader: 验证集data loader
        :return: loss 结果
        """
        return self._train_or_evaluate(phrase=Trainer._EVALUATE,
                                       data_loader=validation_data_loader)

    def _is_serialize_empty(self):
        """
        判断 serialize dir 是否是空的，会忽略 隐藏文件
        :return: True: 空文件夹; False: 非空文件夹
        """
        if not os.path.isdir(self._serialize_dir):
            raise RuntimeError(f"保存路径是无效的路径: {self._serialize_dir} ")

        is_empty = True
        for name in os.listdir(self._serialize_dir):
            if not name.startswith("."):
                is_empty = False
                break
        return is_empty

    def train(self, train_data_loader: DataLoader,
              validation_data_loader: DataLoader) -> None:
        if not self._is_serialize_empty():
            raise RuntimeError(f"新训练，请清空保存文件件: {self._serialize_dir}")

        return self._train(train_data_loader=train_data_loader,
                           validation_data_loader=validation_data_loader)

    def _train(self,
               train_data_loader: DataLoader,
               validation_data_loader: DataLoader) -> None:
        """
        模型训练
        :param train_data_loader: 训练集 data loader
        :param validation_data_loader: 验证集 data loader
        :return:
        """

        if self._current_epoch is None:
            start_epoch = 1
        else:
            start_epoch = self._current_epoch

        for epoch in range(start_epoch, self._num_epoch + 1):

            self._current_epoch = epoch

            logging.info(f"Start train epoch: {self._current_epoch}")

            train_loss = self._train_or_evaluate(phrase=Trainer._TRAIN,
                                                 data_loader=train_data_loader)

            # 输出metrics
            train_metric_dict, train_target_metric = self._metrics.metric
            logging.info(
                f"Train epoch: {epoch}, loss: {train_loss}, target metric: {train_target_metric.name}:{train_target_metric.value} "
                f"metrics: {json2str(train_metric_dict)}")

            evaluate_loss = self.evaluate(validation_data_loader=validation_data_loader)
            validation_metric_dict, validation_target_metric = self._metrics.metric

            self._metric_tracker.add_metric(epoch=epoch,
                                            train_metric=train_metric_dict,
                                            train_model_target_metric=train_target_metric,
                                            validation_metric=validation_metric_dict,
                                            validation_model_target_metric=validation_target_metric)
            logging.info(
                f"Evaluate Valid epoch: {epoch}, loss: {evaluate_loss}, "
                f"target metric: {validation_target_metric.name}:{validation_target_metric.value} "
                f"metrics: {json2str(validation_metric_dict)}")

            # 设置 lr scheduler
            # 注意这样设置会有点问题，对于某些 scheduler step 需要参数, 例如: ReduceLROnPlateau
            # 这种情况，暂时不处理, 如果处理需要重新重构，或者简单实用 isinstance ReduceLROnPlateau 来处理
            # 这里简单实用 isinstance 处理。必须指出，ReduceLROnPlateau 的基类是 object, 这也是多少有些问题。

            if self._lr_scheduler is not None:
                if isinstance(self._lr_scheduler, ReduceLROnPlateau):
                    self._lr_scheduler.step(metrics=evaluate_loss, epoch=epoch)
                else:
                    self._lr_scheduler.step(epoch=epoch)

            self.save_checkpoint(epoch=epoch)

            if self._metric_tracker.early_stopping(epoch):
                logging.info(f"Epoch: {epoch}, early stopping!")
                break



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
通过取到最大值的 sequence label 解码

Authors: PanXu
Date:    2020/07/05 11:28:00
"""

from typing import Tuple
import torch

from easytext.utils import bio as BIO
from easytext.data import LabelVocabulary
from .label_index_decoder import LabelIndexDecoder


class SequenceMaxLabelIndexDecoder(LabelIndexDecoder):
    """
    对于 sequence logits, shape: (batch_size, seq_len, num_label), 使用 max 进行 在每一个
    timestep 上进行 decode, 得到 label index.
    """

    def __init__(self, label_vocabulary: LabelVocabulary):
        """
        初始化
        :param label_vocabulary: label 词汇表
        """
        self._label_vocabulary = label_vocabulary

    def __call__(self, logits: torch.Tensor, mask: torch.ByteTensor) -> torch.LongTensor:
        """
        对于 sequence logits, shape: (batch_size, seq_len, num_label), 使用 max 进行 在每一个
        timestep 上进行 decode, 得到 label index.
        :param logits: shape: (batch_size, seq_len, num_label)
        :param mask: shape: (bath_size, seq_len), 存储的是 0 或 1
        :return: 解码后的 label index, shape: (batch_size, seq_len), 注意这是有padding_index 的结果，
        需要使用 mask 来提取实际的 label index.
        """
        if logits.dim() != 3:
            raise RuntimeError(f"logits shape 错误, 应该是 (B, seq_len, num_label), "
                               f"而现在是 {logits.shape}")

        if (mask is not None) and (mask.dim() != 2):
            raise RuntimeError(f"mask shape 错误, 应该是 (B, seq_len), "
                               f"而现在是 {mask.shape}")

        batch = logits.size(0)
        max_sequence_length = logits.size(1)

        # mask shape: (B, seq_len)
        if mask is None:
            mask = torch.ones(size=(logits.shape[0], logits.shape[1]),
                              dtype=torch.long)

        sequence_length = mask.sum(dim=-1).tolist()

        batch_indices = list()
        for i in range(batch):
            sequence_labels, sequence_label_indices = BIO.decode_one_sequence_logits_to_label(
                sequence_logits=logits[i, :sequence_length[i]],
                vocabulary=self._label_vocabulary)

            padding_indices = [self._label_vocabulary.padding_index] * (max_sequence_length - sequence_length[i])
            sequence_label_indices.extend(padding_indices)

            batch_indices.append(sequence_label_indices)

        batch_indices = torch.tensor(batch_indices, dtype=torch.long)
        return batch_indices



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
acc 指标

Authors: panxu(panxu@baidu.com)
Date:    2020/05/30 07:36:00
"""
from typing import Dict

import torch

from .metric import Metric


class AccMetric(Metric):
    """
    Acc Metric
    """

    ACC = "acc"

    def __init__(self):
        super().__init__()
        self._num_true = 0
        self._num_total = 0

    def __call__(self,
                 prediction_labels: torch.Tensor,
                 gold_labels: torch.Tensor, mask: torch.LongTensor) -> Dict:
        """
        Acc metric 计算
        :param prediction_labels: 预测的标签
        :param gold_labels: gold 标签
        :param mask:
        :return:
        """

        if mask is not None:
            raise RuntimeError("对于 Acc 来说, mask 应该为 None")

        prediction_labels, gold_labels = prediction_labels.detach().cpu(), gold_labels.detach().cpu()

        num_true = (prediction_labels == gold_labels).sum().item()
        num_total = gold_labels.size(0)

        self._num_true += num_true
        self._num_total += num_total

        acc = AccMetric._compute(num_true=num_true, num_total=num_total)
        return {AccMetric.ACC: acc}

    @staticmethod
    def _compute(num_true: int, num_total: int):
        return num_true / (float(num_total) + 1e-10)

    @property
    def metric(self) -> Dict:
        acc = AccMetric._compute(self._num_true, self._num_total)

        return {AccMetric.ACC: acc}

    def reset(self):
        self._num_true = 0
        self._num_total = 0
        return self




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
Glove 的词向量

Authors: panxu(panxu@baidu.com)
Date:    2020/06/25 09:57:00
"""

from .general_pretrained_word_embedding_loader import GeneralPretrainedWordEmbeddingLoader


class GloveLoader(GeneralPretrainedWordEmbeddingLoader):
    """
    Glove Word Embedding Loader.
    glove.6B.50d.txt
    glove.6B.100d.txt
    glove.840B.300d.txt
    """

    def __init__(self, embedding_dim: int, pretrained_file_path: str, max_size: int = None):
        """
        初始化 glove embedding loader
        :param embedding_dim: embedding 维度
        :param pretrained_file_path: 预训练的文件路径, 一般是指:
        glove.6B.50d.txt glove.6B.100d.txt glove.840B.300d.txt ...
        :param max_size: 载入的最大向量数。因为某些词向量可能很大，比如 glove.840B.300d.txt 有 5G 大小，
        通过该参数进行限制，在比如测试阶段是非常有益的，或者一些内存小的机器上。
        """
        super().__init__(embedding_dim=embedding_dim,
                         pretrained_file_path=pretrained_file_path,
                         encoding="utf-8",
                         separator=" ",
                         skip_num_line=0,
                         max_size=max_size)



# coding:utf-8
"""
json 工具
"""
import torch
import json
from datetime import datetime, date

from easytext.utils import dt_util


class _JsonEncoder(json.JSONEncoder):
    """
    全局的json encoder
    """

    def default(self, o):
        """
        json encoder执行的地方
        :param o: object
        :return:
        """

        if isinstance(o, datetime):
            return o.strftime(dt_util.FORMATE_1)
        elif isinstance(o, date):
            return o.strftime(dt_util.FORMATE_3)
        elif isinstance(o, torch.Tensor):
            if o.dim() == 0:
                return o.item()
            else:
                return o.tolist()
        else:
            pass

        return o.__dict__


def json2str(json_obj: json, cls: json.JSONEncoder = _JsonEncoder, indent: int = None) -> str:
    """
    obj转换成json字符串
    :param json_obj: 转换的对象
    :param cls: json encoder
    :param indent: 对json进行美化. index=4 表示按照4个空格进行缩进
    :return: 字符串
    """
    return json.dumps(json_obj, cls=cls, ensure_ascii=False, indent=indent)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
安装脚本

Authors: panxu
Date:    2020/06/08 15:44:00
"""

from setuptools import setup
from setuptools import find_packages


setup(
    name="easytext-nlp",
    version="0.0.2",
    author="Pan Xu",
    author_email="cjopengler@163.com",
    description="Make it easy to train and metric NLP model.",
    url="https://github.com/cjopengler/easytext",
    license="MIT",
    packages=find_packages(include=("easytext","easytext.*")),   # 指定需要安装的模块

    install_requires=["torch>=1.4.0"],
    python_requires='>=3.6',

    classifiers=[
        # How mature is this project? Common values are
        # 3 - Alpha
        # 4 - Beta
        # 5 - Production/Stable
        'Development Status :: 3 - Alpha',
        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',

        "Operating System :: OS Independent",
    ],
    zip_safe=False
)




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 cnn seq2vec

Authors: PanXu
Date:    2020/07/08 20:10:00
"""
import torch

from easytext.modules.seq2vec import CnnSeq2Vec
from easytext.tests import ASSERT


def test_cnn_seq2vec():
    """
    测试 cnn seq2vec
    :return:
    """

    encoder = CnnSeq2Vec(embedding_dim=2, num_filters=1, kernel_sizes=(1, 2))

    for name, parameter in encoder.named_parameters():
        parameter.data.fill_(1.)

    tokens = torch.FloatTensor([[[0.7, 0.8], [0.1, 1.5]]])
    vector = encoder(tokens=tokens, mask=None)
    vector = vector.view(-1).tolist()

    expect = torch.tensor([[0.1 + 1.5 + 1., 0.7 + 0.8 + 0.1 + 1.5 + 1.]]).view(-1).tolist()

    ASSERT.assertEqual(len(expect), len(vector))
    for i in range(len(vector)):
        ASSERT.assertAlmostEqual(expect[i], vector[i])


def test_cnn_seq2vec_with_mask():
    """
    测试带有 mask 的 cnn seq2vec
    :return:
    """

    encoder = CnnSeq2Vec(embedding_dim=2, num_filters=1, kernel_sizes=(1, 2))

    for name, parameter in encoder.named_parameters():
        parameter.data.fill_(1.)

    tokens = torch.FloatTensor([[[0.7, 0.8], [0.1, 1.5]]])
    mask = torch.ByteTensor([[1, 0]])
    vector = encoder(tokens=tokens, mask=mask)
    vector = vector.view(-1).tolist()

    expect = torch.tensor([[0.7 + 0.8 + 1., 0.7 + 0.8 + 0. + 0. + 1.]]).view(-1).tolist()

    ASSERT.assertEqual(len(expect), len(vector))
    for i in range(len(vector)):
        ASSERT.assertAlmostEqual(expect[i], vector[i])


def test_cnn_seq2vec_output_dim():
    """
    测试 cnn 输出维度
    :return:
    """
    kernel_size = (1, 2, 3, 4, 5)
    encoder = CnnSeq2Vec(embedding_dim=7,
                         num_filters=13,
                         kernel_sizes=kernel_size)

    tokens = torch.rand(4, 8, 7)
    vector = encoder(tokens=tokens, mask=None)
    expect = (4, 13 * len(kernel_size))

    ASSERT.assertEqual(expect, vector.size())



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
测试 保存和载入 checkpoint

Authors: panxu(panxu@baidu.com)
Date:    2020/05/29 11:11:00
"""
import json
import os
import shutil
import logging
from typing import Iterable, List, Dict, Tuple

import torch
from torch import Tensor
from torch.utils.data import Dataset, DataLoader

from easytext.data import Instance, LabelVocabulary
from easytext.data.model_collate import ModelInputs
from easytext.data.model_collate import ModelCollate
from easytext.model import Model
from easytext.model import ModelOutputs
from easytext.loss import Loss
from easytext.optimizer import OptimizerFactory
from easytext.trainer import Trainer
from easytext.metrics import AccMetric, ModelMetricAdapter, ModelTargetMetric
from easytext.utils import log_util
from easytext.utils.json_util import json2str
from easytext.label_decoder import ModelLabelDecoder
from easytext.label_decoder import MaxLabelIndexDecoder

from easytext.tests import ASSERT
from easytext.tests import ROOT_PATH


log_util.config()


class _DemoDataset(Dataset):
    """
    测试用的数据集
    """

    def __init__(self):
        self._instances: List[Instance] = list()

        max_num = 100
        for i in range(0, 50):
            instance1 = Instance()
            instance1["x"] = i

            instance2 = Instance()
            instance2["x"] = max_num - i

            self._instances.append(instance1)
            self._instances.append(instance2)

    def __len__(self):
        return len(self._instances)

    def __getitem__(self, index):
        return self._instances[index]


class _DemoCollate(ModelCollate):
    """
    测试用的 collate
    """

    def __call__(self, instances: Iterable[Instance]) -> ModelInputs:

        x = list()
        labels = list()
        for instance in instances:

            x_data = instance["x"]
            x.append(torch.tensor([x_data], dtype=torch.float))

            if x_data - 50 > 0:
                labels.append(1)
            else:
                labels.append(0)

        x = torch.stack(x)

        batch_size = x.size(0)
        ASSERT.assertEqual(x.dim(), 2)
        ASSERT.assertListEqual([batch_size, 1], [x.size(0), x.size(1)])

        labels = torch.tensor(labels)
        ASSERT.assertEqual(labels.dim(), 1)
        ASSERT.assertEqual(batch_size, labels.size(0))

        model_inputs = ModelInputs(batch_size=batch_size,
                                   model_inputs={"x": x},
                                   labels=labels)

        return model_inputs


class _DemoOutputs(ModelOutputs):

    def __init__(self, logits: torch.Tensor):
        super().__init__(logits=logits)


class _DemoLabelDecoder(ModelLabelDecoder):

    def __init__(self):
        super().__init__()
        self._label_index_decoder = MaxLabelIndexDecoder()

    def decode_label_index(self, model_outputs: ModelOutputs) -> torch.LongTensor:
        model_outputs: _DemoOutputs = model_outputs
        return self._label_index_decoder(model_outputs.logits)

    def decode_label(self, model_outputs: ModelOutputs, label_indices: torch.LongTensor) -> List:
        return [label_index.item() for label_index in label_indices]


class _DemoMetric(ModelMetricAdapter):

    def __init__(self):
        super().__init__()
        self._acc = AccMetric()
        self._label_decoder = _DemoLabelDecoder()

    def __call__(self, model_outputs: _DemoOutputs, golden_labels: Tensor) -> Tuple[Dict, ModelTargetMetric]:
        model_outputs: _DemoOutputs = model_outputs
        label_indices = self._label_decoder.decode_label_index(model_outputs=model_outputs)
        acc = self._acc(prediction_labels=label_indices, gold_labels=golden_labels, mask=None)

        target = ModelTargetMetric(AccMetric.ACC, acc[AccMetric.ACC])
        return acc, target

    @property
    def metric(self) -> Tuple[Dict, ModelTargetMetric]:
        acc = self._acc.metric

        target = ModelTargetMetric(AccMetric.ACC, acc[AccMetric.ACC])
        return acc, target

    def reset(self) -> "_DemoMetric":
        self._acc.reset()
        return self


class _DemoOptimizerFactory(OptimizerFactory):

    def create(self, model: Model) -> "Optimizer":
        return torch.optim.Adam(params=model.parameters(), lr=1e-1)


class ModelDemo(Model):

    def __init__(self):
        super().__init__()
        self.feed_forward = torch.nn.Linear(in_features=1, out_features=2)

        self.reset_parameters()

    def reset_parameters(self):
        pass

    def forward(self, x: torch.Tensor) -> _DemoOutputs:

        # print(self.feed_forward.weight, self.feed_forward.bias)
        logits = self.feed_forward(x)

        demo_outputs = _DemoOutputs(logits=logits)
        return demo_outputs


class _DemoLoss(Loss):

    def __init__(self):
        super().__init__()
        self._loss = torch.nn.CrossEntropyLoss()

    def __call__(self, model_outputs: ModelOutputs, golden_label: torch.Tensor) -> torch.Tensor:
        model_outputs: _DemoOutputs = model_outputs

        return self._loss(model_outputs.logits, golden_label)


def _run_train(cuda_devices: List[str] = None):
    serialize_dir = os.path.join(ROOT_PATH, "data/easytext/tests/trainer/save_and_load")

    if os.path.isdir(serialize_dir):
        shutil.rmtree(serialize_dir)

    os.makedirs(serialize_dir)

    model = ModelDemo()

    optimizer_factory = _DemoOptimizerFactory()

    loss = _DemoLoss()
    metric = _DemoMetric()

    trainer = Trainer(num_epoch=100,
                      model=model,
                      loss=loss,
                      metrics=metric,
                      optimizer_factory=optimizer_factory,
                      serialize_dir=serialize_dir,
                      patient=20,
                      num_check_point_keep=25,
                      cuda_devices=cuda_devices
                      )

    train_dataset = _DemoDataset()

    train_data_loader = DataLoader(dataset=train_dataset,
                                   collate_fn=_DemoCollate(),
                                   batch_size=200,
                                   shuffle=False,
                                   num_workers=0)

    validation_data_loader = DataLoader(dataset=train_dataset,
                                        collate_fn=_DemoCollate(),
                                        batch_size=200,
                                        shuffle=False,
                                        num_workers=0)

    trainer.train(train_data_loader=train_data_loader,
                  validation_data_loader=validation_data_loader)

    expect_model_state_dict = json.loads(json2str(trainer.model.state_dict()))
    expect_optimizer_state_dict = json.loads(json2str(trainer.optimizer.state_dict()))
    expect_current_epoch = trainer.current_epoch
    expect_num_epoch = trainer.num_epoch
    expect_metric = trainer.metrics.metric[0]
    expect_metric_tracker = json.loads(json2str(trainer.metric_tracker))

    trainer.load_checkpoint(serialize_dir=serialize_dir)

    loaded_model_state_dict = json.loads(json2str(trainer.model.state_dict()))
    loaded_optimizer_state_dict = json.loads(json2str(trainer.optimizer.state_dict()))
    current_epoch = trainer.current_epoch
    num_epoch = trainer.num_epoch
    metric = trainer.metrics.metric[0]
    metric_tracker = json.loads(json2str(trainer.metric_tracker))

    ASSERT.assertDictEqual(expect_model_state_dict, loaded_model_state_dict)
    ASSERT.assertDictEqual(expect_optimizer_state_dict, loaded_optimizer_state_dict)
    ASSERT.assertEqual(expect_current_epoch, current_epoch)
    ASSERT.assertEqual(expect_num_epoch, num_epoch)
    ASSERT.assertDictEqual(expect_metric, metric)
    ASSERT.assertDictEqual(expect_metric_tracker, metric_tracker)


def test_trainer_save_and_load_cpu():
    """
    测试  trainer 保存和载入
    :return:
    """
    _run_train()


def test_trainer_save_load_gpu():

    if torch.cuda.is_available():
        cuda_devices = ["cuda:0"]
        _run_train(cuda_devices=cuda_devices)
    else:
        logging.warning("由于没有GPU，忽略这个case测试")





#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
text cnn

Authors: PanXu
Date:    2020/07/07 20:49:00
"""
from typing import List, Union

import torch
from torch.nn import Module
from torch.nn import Conv1d
from torch.nn import ModuleList
from torch.nn import ReLU
from torch.nn import MaxPool1d


class CnnSeq2Vec(Module):
    """
    text cnn 编码器，将文本经过cnn
    完成的是 sequence 到 vector 转化
    """

    def __init__(self,
                 embedding_dim: int,
                 num_filters: int,
                 kernel_sizes: List[int]):
        """
        初始化
        :param embedding_dim: 输入的维度
        :param num_filters: filter 数量，也就是输出维度
        :param kernel_sizes: kernel size
        """
        super().__init__()
        self.embedding_dim = embedding_dim
        self.num_filters = num_filters
        self.cnn_layers = ModuleList([Conv1d(in_channels=embedding_dim,
                                             out_channels=num_filters,
                                             kernel_size=kernel_size) for kernel_size in kernel_sizes])

        self.activtion = ReLU()

    def forward(self, tokens: torch.LongTensor, mask: Union[None, torch.ByteTensor]) -> torch.Tensor:
        """
        执行模型。对多个 kernel size 最终会将每一个 kernel 输出的向量, concat 在一起。
        pooling 使用的 max pooling.
        :param tokens: 输入的token 序列, shape: (batch_size, seq_len, embedding_dim)
        :param mask: mask
        :return: cnn 编码向量, shape: (batch_size, num_filter * len(kernel_sizes))
        """

        assert tokens.dim() == 3, f"tokens.dim: {tokens.dim()} 与 shape: (batch_size, seq_len, embedding_dim) 不匹配"

        if mask is not None:
            assert mask.dim() == 2, f"mask.dim: {mask.dim()} 与 shape: (batch_size, seq_len) 不匹配"

            # 将 mask 的 token 清零，避免影响 cnn
            tokens = tokens * mask.unsqueeze(dim=-1).float()

        # 将 1 和 2 转置, 转置后 shape: (batch_size, embedding_dim, seq_len)
        tokens = torch.transpose(tokens, 1, 2)

        # 每一个 cnn_vector_i: (batch_size, embedding_dim, new_seq_len_i)
        # 注意不同 kernel_size 的 cnn, 产生的 new_seq_len 长度是不同的 所以这里用下标 i 来表示.
        cnn_vectors = [self.activtion(cnn(tokens)) for cnn in self.cnn_layers]

        assert cnn_vectors[0].dim() == 3, \
            f"cnn_vectors.dim: {cnn_vectors[0].dim()} 与 shape: (batch_size, num_filter, new_seq_len) 不匹配"
        assert cnn_vectors[0].size(1) == self.num_filters

        # max pooling, 直接使用 max，而不是使用 MaxPool1D, max 更方便，MaxPool1D 需要设置 kernel size 为 seq_len.
        max_pooled_cnn_vectors = [cnn_vector.max(dim=-1)[0] for cnn_vector in cnn_vectors]
        assert max_pooled_cnn_vectors[0].dim() == 2, \
            f"max_pooled_cnn_vectors.dim: {max_pooled_cnn_vectors[0].dim()} 与 shape: (batch_size, num_filter) 不匹配"

        # 最后 max_pooled_cnn_vectors concat 在一起
        vector = \
            torch.cat(max_pooled_cnn_vectors, dim=-1) if len(max_pooled_cnn_vectors) > 1 else max_pooled_cnn_vectors[0]

        return vector




